// For licensing, see: http://hydrogen.knownspace.org/license.html
package org.datamanager.kernel;

import org.datamanager.constraint.EventConstraint;

import org.datamanager.event.DataManagerEvent;

import org.datamanager.exception.DataManagerException;

import org.datamanager.kernel.DataManagerWorker;

import org.datamanager.tools.Debug;

/** $Id: AbstractEventConstraint.java,v 1.6 2002/10/08 02:08:37 alllee Exp $
 * This class is an abstract EventConstraint class. It implements the
 * EventConstraint interface located in org.datamanager.constraint,
 * and provides a public describes() method, which is exactly the
 * same as the public abstract accepts() method in EventConstriant.
 * Every EventConstraint should ideally extend this class instead of
 * implementing EventConstraint themselves. By doing so the constraint
 * will gain extra power while working with Entities in the kernel.
 * For example, a when a constraint is working on an Entity, calling
 * methods such as getValue() and getAttributes(), it should not
 * cause DataManagerEvents to be fired. If the constriant is a
 * subclass of AbstractConstraint then no events will be fired,
 * otherwise events will.
 * 
 * If a constraint has already been written to implement
 * EventConstraint it can easily be modified to use this class
 * instead simply by changing implements EventConstraint to
 * extends AbstractEventConstraint and renaming the accepts() method
 * to describes().
 * 
 * @author Matthew Farrellee
 *
 * @see org.datamanager.constraint.EntityConstraint
 * @see org.datamanager.event.DataManagerEvent
 * @version $Revision: 1.6 $
 * 
 * 
 */
public abstract class AbstractEventConstraint
    implements EventConstraint
    {
    public static final String DEBUG_ACCEPTS_MESSAGE = 
            "Non-kernel thread attempted to invoke accepts(...) on DataManagerEvent ";

    /**
     * The accepts() method required by the EventConstraint interface.
     * This class implements this method and makes it final, so
     * no subclasses can override the work that is done by this
     * class. Eventually this accepts() method will call the
     * describes() method that should be implemented by subclasses.
     *
     * @param event An event to check.
     *
     * @return Whether the event is acceptable or not.
     */
    public final boolean accepts(DataManagerEvent event)
                          throws DataManagerException
        {
        // assume constraint is initially invalid 
        boolean isConstraintValid = false;

        // If the event is null then it cannot be matched.
        if (event != null)
            {
            // Make the DataManagerWorker not event-generating.
            // If the thread is not an DataManagerWorker
            // there should be a DataManagerException.
            try
                {
                // Note: Normally this code could be considered having
                // a bug since if two threads were executing in this
                // code at once the following could happen:
                // ThreadA gets oldValue = eventGenerating = true
                // ThreadA sets eventGenerating = false
                // ThreadB gets oldValue = eventGenerating = false
                // ThreadA sets eventGenerating = oldValue = true
                // ThreadB sets eventGenerating = oldValue = false
                // The end result would be a thread with invalid state.
                // However, since eventGenerating is specific to threads
                // ThreadA and ThreadB would not actually be sharing the
                // same generateEvents state and therefore this is not
                // a problem, but if for some reason in the future
                // the generateEvents state is not associated with
                // threads this code could become a problem.
                DataManagerWorker thread = (DataManagerWorker) Thread.currentThread();
                boolean oldValue = thread.isEventGenerating();

                thread.generateEvents(false);

                isConstraintValid = describes(event);

                // Make the DataManagerWorker back to the way it
                // was.
                thread.generateEvents(oldValue);
                }
            catch (ClassCastException exception)
                {
                // This should really never happen. If it has
                // then it means that someone is running a thread
                // of their own that is not valid.
                // The thread should really be killed at this point,
                // but maybe it would be better just to report the
                // rogue thread to some thread manager.
                Debug.writeLogMessage(DEBUG_ACCEPTS_MESSAGE + event + 
                                      "\n\t Thread was: " + 
                                      Thread.currentThread());

                // TODO: Destroy thread in some thread-safe manner.
                // in some manner such as:
                // DataManager.FunkUp(Thread.currentThread());

                /*
                Debug.assert("Non-kernel thread running in the kernel: " + 
                Thread.currentThread());
                */
                }
            }

        // This constraint will just return false if the thread
        // is invalid.
        return isConstraintValid;
        }

    /**
     * The method required by this class. This method has the
     * exact same semantics as the interface EventConstraint's
     * accepts() method.
     *
     * @param event An entity to check.
     *
     * @return Whether the event is acceptable or not.
     */
    protected abstract boolean describes(DataManagerEvent event)
                                  throws DataManagerException;
    }