// For licensing, see: http://hydrogen.knownspace.org/license.html
//
//Modified (10/06/2002): nab
//1) Modified the listeners instance variable to be a Map interface
//instaed of being a concrete Hashtable
//2) Modified the contraints holder which used to be a vector for each
//entry in the hashtable, to use the List interface instead
//3) Modified the Vector to be of type ArrayList
//4) Converted the enumerations in fireEvents to be iterators
//5) Extract method to move the logic to check the individual listeners
//constraints and fire the event from fireEvent to
//fireEventsToHandlerAccordingToConstraints
//6) Extract method to to move the iteration logic and casting of listeners
//from fireEvent to a new method named fireEventHandlers
//7) Used replace temp with query and decompose conditional to move the check
//to see if the current thread is an event generating thread to a new method
//named shouldGenerateEvents
//8) Renamed private instance variable listeners to be eventListeners
//9) Made private accessor method getEventListeners so that we are always
//calling a method to access the object.
//10) Used decompose conditional to make the check of a constraint for a given
//handler more readable in a method named shouldFireEventToHandler
//11) Eemoved small size initialization to go with java defaults since we are
//not supposed to worry about performance at this phase in the ball game
//12) Fixed up the javadoc and removed where no longer needed because of
//refactorings performed on the class
//13) Replaced the Hashtable with a HashMap
//14) Changed Debug.assert() to Debug.writeLogMessage for doing logging.
//
//Modified (10/10/2002): maa
//1) Cleaned up the JavaDoc a bit.
//2) Wrote log message when DataManagerException is caught in fireEventsToHandler()
//
package org.datamanager.kernel;

import java.io.Serializable;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.datamanager.constraint.EventConstraint;

import org.datamanager.event.DataManagerEvent;

import org.datamanager.exception.DataManagerException;

import org.datamanager.tools.Debug;

/**
 * There are going to be many EventGenerators in the system,
 * and all of them will have the same core responsibilities.
 * This class is an attempt to give EventGenerators a way to
 * perform those responsibilities. This class provides methods
 * that can be delegated to by EventGenerators.
 *
 * @author Matthew Farrellee
 *
 * @see org.datamanager.kernel.EventGenerator
 */
class EventGeneratorAssistant
    implements Serializable
    {
    /**
     * EventHandlers and EventConstraints are stored
     * in this Map with the EventHandler as the
     * key and an ArrayList of EventConstraints that the handler has been
     * subscribed with as the value.
     */
    private Map eventListeners;

    /**
     * Create a new EventGeneratorAssistant.
     */
    protected EventGeneratorAssistant()
        {
        eventListeners = new HashMap();
        }

    /**
     * This method allows EventGenerators to add an EventHandler and EventConstraint.
     * This should be used within the EventGenerator's subscribe() method.
     *
     * @param handler EventHandler to subscribe.
     * @param constraint EventConstraint to filter with.
     */
    protected synchronized void addEventHandler(EventHandler handler, 
                                                EventConstraint constraint)
        {
        if (getEventListeners().containsKey(handler))
            {
            ((List) getEventListeners().get(handler)).add(constraint);
            }
        else
            {
            List constraints = new ArrayList();
            constraints.add(constraint);
            getEventListeners().put(handler, constraints);
            }
        }

    /**
     * Remove an EventHandler. This method provides access
     * for removing EventHandlers. It should be used within
     * the EventGenerator's unsubscribe() method.
     *
     * @param handler The EventHandler to remove.
     */
    protected synchronized void removeEventHandler(EventHandler handler)
        {
        getEventListeners().remove(handler);
        }

    /**
     * Remove a specific constraint associated with the given
     * EventHandler. This method provides access for removing
     * EventHandler's specific constraints. Unlike the single
     * argument version of this method. This method will remove
     * only one constraint that the EventHandler is subscribed
     * with.
     *
     * @param handler EventHandler to remove a constraint for.
     * @param constraint The constraint to remove.
     */
    protected synchronized void removeEventHandler(EventHandler handler, 
                                                   EventConstraint constraint)
        {
        if (getEventListeners().containsKey(handler))
            {
            List constraints = (List) getEventListeners().get(handler);
            constraints.remove(constraint);

            if (constraints.isEmpty())
                {
                removeEventHandler(handler);
                }
            }
        }

    /**
     * Fire the given event to all interested EventHandlers.
     * This method provides a way for DataManagerEvents to be
     * sent to all EventHandlers subscribed for them. This
     * method should be used in an EventGenerator whenever an
     * event needs to be fired.
     * Note: Currently a single EventHandler may receive more
     * than one event. This will happen if it is subscribed
     * with more than one constraint that accepts the event.
     *
     * @param event The event to fire.
     */
    protected synchronized void fireEvent(DataManagerEvent event)
        {
        try
            {
            if (shouldGenerateEvents())
                {
                fireEventHandlers(event);
                }
            }
        catch (ClassCastException exception)
            {
            Debug.writeLogMessage("illegal thread running in the system: " + 
                                  Thread.currentThread());
            }
        }

    /**
     * check whether the current DataManagerWorker thread should generate
     * events by calling its isEventGenerating method
     *
     * @return true if events should be generated false otherwise
     */
    private synchronized boolean shouldGenerateEvents()
        {
        return ((DataManagerWorker) Thread.currentThread()).isEventGenerating();
        }

    /**
     * Go through all of the event listeners and fire the event to each one
     *
     * @param event The event to fire.
     */
    private synchronized void fireEventHandlers(DataManagerEvent event)
        {
        Iterator handlers = getEventListeners().keySet().iterator();

        while (handlers.hasNext())
            {
            fireEventsToHandler(event, (EventHandler) handlers.next());
            }
        }

    /**
     * For a given event handler check all of its constraints to see if
     * the event should be fired and if so then create a new EventHandlerThread
     * to carry out the passing of the event to the handler
     *
     * @param event The event to fire.
     * @param handler The event handler to fire the event at
     */
    private synchronized void fireEventsToHandler(DataManagerEvent event, 
                                                  EventHandler handler)
        {
        Iterator constraints = ((List) getEventListeners().get(handler)).iterator();

        while (constraints.hasNext())
            {
            if (shouldFireEventToHandler(event, 
                                         (EventConstraint) constraints.next()))
                {
                new EventHandlerThread(handler, event).start();
                }
            }
        }

    /**
     * Check whether or not an event should be fired based on an event and eventConstraint
     *
     * @param event The event to fire.
     * @param eventConstraint The EventConstraint to check the event against
     * @return true if the event should be fired to the handler false otherwise
     */
    private boolean shouldFireEventToHandler(DataManagerEvent event, 
                                             EventConstraint eventConstraint)
        {
        try
            {
            return (eventConstraint.accepts(event));
            }
        catch (DataManagerException dataManagerException)
            {
            Debug.writeLogMessage(
                    "DataManagerException was thrown indicating event should not be fired to handler!");

            return false;
            }
        }

    /**
     * Simple getter method for our event listeners map
     *
     * @return Map of eventListeners
     */
    protected Map getEventListeners()
        {
        return eventListeners;
        }
    }