// For licensing, see: http://hydrogen.knownspace.org/license.html
//
//IdentifierManager.java
//Created: 11/19/1999 16:44
//
//Modified (10/06/2002): maa
//1) Cleaned up JavaDoc to adhere to standards and eliminate conflicts.
//2) Cleaned up equals method (redundant null check after instanceof check).
//3) Used replace temp with query to eliminate temp index variable in SimpleIdentifier constructor.
//4) Renamed identifier and extra instance variables of SimpleIdentifier to stringSection and longSection respectively.
//5) Renamed format temp variable in generateIdentifier() to formattedDateString for clarity.
///////////////////////////////////////////////////////////////////////////////
// 2002-10-11 cam
// 1. Cleaned up JavaDoc.
// 2. Further simplified the SimpleIdentifier.equals() method so no temp
//    variables or conditionals are needed.
///////////////////////////////////////////////////////////////////////////////
package org.datamanager.kernel;

import java.io.Serializable;

import java.text.SimpleDateFormat;

import java.util.Date;

import org.datamanager.kernel.Identifier;

/**
 * The IdentifierManager class is a source of unique Identifiers.
 *
 * @author Benjamin Keil
 * @author Matthew Farrellee
 */
class IdentifierManager
    implements Serializable
    {
    /** The only instance of the IdentifierManager class. */
    private static IdentifierManager instance = new IdentifierManager();

    /**
     * The default constructor. It is private because IdentifierManager
     * is a singleton.
     */
    private IdentifierManager()
        {
        }

    /**
     * Returns the Singleton instance of the IdentifierManager.
     *
     * @return The instance of IdentifierManager.
     */
    public static IdentifierManager getInstance()
        {
        return instance;
        }

    /**
     * Returns the next available unique identifier.
     *
     * @return A new identifier.
     */
    public static Identifier getIdentifier()
        {
        return SimpleIdentifier.generateIdentifier();
        }

    /**
     * The SimpleIdentifier is an inner class that is responsible for generating unique
     * Identifiers, using a combination of the date and a long.
     *
     * @author Benjamin Keil
     * @author Matthew Farrellee
     */
    private static class SimpleIdentifier
        implements Identifier
        {
        /** The last Identifier that was created. */
        protected static SimpleIdentifier lastIdentifier = 
                new SimpleIdentifier("0:0");
        /** DateFormat to assist in creating unique Identifiers. */
        protected static final SimpleDateFormat dateFormat = 
                new SimpleDateFormat("yyyy.MM.dd.hh.mm.ss.SSSS.z");
        /** The String value of the Identifier. */
        protected final String stringSection;
        /** A number to help distinguish between Identifiers. */
        protected final long longSection;

        /**
         * Constructs a unique Identifier.
         *
         * @param stringSection A String version of the date, formatted according to the
         *                      dateFormat constant.
         * @param longSection The current incremented value of the long numeric that ensures
         *                    that Indentifiers are unique.
         */
        public SimpleIdentifier(String stringSection, long longSection)
            {
            this.stringSection = stringSection;
            this.longSection = longSection;
            }

        /**
         * Creates a new SimpleIdentified from the given String.
         *
         * @param string The String value of the Identifier.
         */
        public SimpleIdentifier(String identifierString)
            {
            this.stringSection = identifierString.substring(0, 
                                                            identifierString.indexOf(
                                                                    ":"));
            this.longSection = Long.parseLong(identifierString.substring(identifierString.indexOf(
                                                                                 ":") + 1, 
                                                                         identifierString.length()));
            }

        /**
         * Gets the String part of this Identifier.
         *
         * @return The String part of the Identifier
         */
        protected String getStringSection()
            {
            return this.stringSection;
            }

        /**
         * Gets the long part of this Identifier.
         *
         * @return The long part of the Indentifier.
         */
        protected long getLongSection()
            {
            return this.longSection;
            }

        public int hashCode()
            {
            return this.toString().hashCode();
            }

        /**
         * Checks to see if the given Object is the same as this Identifier.
         *
         * @param object A candidate for equality.
         * @return A boolean indicating whether the given object and the Identifer are
         *         the same.
         */
        public boolean equals(Object object)
            {
            return (object instanceof SimpleIdentifier) && 
                   object.toString().equals(this.toString());
            }

        /**
         * Gets a String representation of this Identifier.
         *
         * @return A String representation of the Identifier.
         */
        public String toString()
            {
            return stringSection + ":" + longSection;
            }

        /**
         * Generates a unique Identifier.
         *
         * @return A new, unique Identifier.
         */
        protected static Identifier generateIdentifier()
            {
            String formattedDateString = dateFormat.format(new Date());

            long longSection = 0;

            if (lastIdentifier.getStringSection().equals(formattedDateString))
                {
                longSection = lastIdentifier.getLongSection() + 1;
                }

            SimpleIdentifier identifier = new SimpleIdentifier(
                                                  formattedDateString, 
                                                  longSection);

            lastIdentifier = identifier;

            return identifier;
            }
        }
    }