// For licensing, see: http://hydrogen.knownspace.org/license.html
package org.datamanager.persistence.xml;

import java.lang.reflect.*;

import org.datamanager.persistence.xml.fieldtypes.*;
import org.datamanager.persistence.xml.fieldtypes.FieldType;
import org.datamanager.persistence.xml.fieldtypes.TypeMap;

import org.w3c.dom.*;

/**
 * rebuild the object from Xml file.
 * It will use reflection to rebuild the object by using empty constructor. Then It will use
 * relection to set the state of object, which is keep in the xml file on disk.
 *
 * @author xi rao & Jason Lynn Baumgartner
 */

// 10/12/02 nab - modified  added null checking that was breaking allan and ourselves in Pool testing
public class XMLDataRestore
    {
    private ObjectIDMap objectIDMap;
    private TypeMap fieldTypeMap;

    /**
     * Constructor: create an XMLDataRestore instance with ObjectIDMap and fieldTypeMap table
     */
    protected XMLDataRestore()
        {
        objectIDMap = new ObjectIDMap();
        fieldTypeMap = new TypeMap();
        }

    // return the ObjectIDMap of XMLDataRestore
    private ObjectIDMap getObjectIDMap()
        {
        return objectIDMap;
        }

    // return the FieldTypeMap of XMLDataRestore
    private TypeMap getFieldTypeMap()
        {
        return fieldTypeMap;
        }

    /**
     * rebuild object instance, restore its state and return it
     * @param document Document keep the objest state information
     * @return Object instance of rebuild object
     */
    protected Object getRestoredObject(Document document)
        {
        Object restoreObject = null;
        NodeList javaNodeList = document.getElementsByTagName(
                                        XMLDataAccessObject.JAVA);

        if ((javaNodeList != null) && (javaNodeList.item(0) != null))
            {
            NodeList children = javaNodeList.item(0).getChildNodes();

            for (int i = 0; i < children.getLength(); i++)
                {
                Node objectNode = children.item(i);

                if (objectNode.ELEMENT_NODE == Node.ELEMENT_NODE)
                    {
                    restoreObject = getObjectInstance(objectNode);
                    }
                }
            }

        return restoreObject;
        }

    /**
     * rebuild the correspond object instance in the Node(element) and restore the state from xml
     * file
     * @param node Node element of xml
     * @return Object instane
     */
    private Object getObjectInstance(Node node)
        {
        Object instance = null;
        NamedNodeMap attributes = node.getAttributes();

        if (supported(attributes, XMLDataAccessObject.IDREF))
            {
            Node idreferenceNode = attributes.getNamedItem(
                                           XMLDataAccessObject.IDREF);
            instance = getObjectIDMap()
                           .getObject(idreferenceNode.getNodeValue());
            }
        else
            {
            if (supported(attributes, XMLDataAccessObject.CLASS))
                {
                Node classNameNode = attributes.getNamedItem(
                                             XMLDataAccessObject.CLASS);

                try
                    {
                    Class type = Class.forName(classNameNode.getNodeValue());
                    instance = constructObject(type);

                    Node idNode = attributes.getNamedItem(
                                          XMLDataAccessObject.ID);
                    getObjectIDMap().add(idNode.getNodeValue(), instance);
                    addObjectProperties(instance, node);
                    }
                catch (ClassNotFoundException cnfe)
                    {
                    cnfe.printStackTrace();
                    }
                }
            }

        return instance;
        }

    /**
     * rebuild the correspond object instance
     * @param type
     * @return
     */
    private Object constructObject(Class type)
        {
        Object instance = null;

        try
            {
            Constructor constructor = getMinimumConstructor(type);
            constructor.setAccessible(true);
            instance = constructor.newInstance(getConstructorParameter(
                                                       constructor));
            }
        catch (InvocationTargetException ite)
            {
            ite.printStackTrace();
            }
        catch (IllegalAccessException iae)
            {
            iae.printStackTrace();
            }
        catch (InstantiationException ie)
            {
            ie.printStackTrace();
            }
        catch (Exception everything)
            {
            everything.printStackTrace();
            }

        return instance;
        }

    // provide the dummy ojbect to correspond to constructor's parameter type
    private Object[] getConstructorParameter(Constructor constructor)
        {
        int arrayLength = constructor.getParameterTypes().length;
        Object[] parameters = new Object[arrayLength];
        Class[] types = constructor.getParameterTypes();

        for (int i = 0; i < arrayLength; i++)
            {
            parameters[i] = fieldTypeMap.getDummyObject(types[i].getName());
            }

        return parameters;
        }

    /**
     * return the constructor of the object which is state in the xml file
     * Fixme(10/10/2002): we better require the client provide the empty constructor. If we require the client
     * provide the empty, return the minimum constructor is useless
     * @param type
     * @return
     */
    private Constructor getMinimumConstructor(Class type)
        {
        int paramCount = 100;
        Constructor minimumConstructor = null;

        //Constructor[] constructors = type.getConstructors();
        Constructor[] constructors = type.getDeclaredConstructors();

        for (int i = 0; i < constructors.length; i++)
            {
            Constructor constructor = constructors[i];
            constructor.setAccessible(true);

            if (constructor.getParameterTypes().length < paramCount)
                {
                paramCount = constructor.getParameterTypes().length;
                minimumConstructor = constructor;
                }
            }

        return minimumConstructor;
        }

    /**
     * add state to object. The state is keep in the xml file and store on disk
     * @param instance Object instace will be add the state
     * @param parent Node keep the state in xml format (Element)
     */
    private void addObjectProperties(Object instance, Node parent)
        {
        NodeList voidList = parent.getChildNodes();

        for (int i = 0; i < voidList.getLength(); i++)
            {
            Node voidNode = voidList.item(i);
            NamedNodeMap voidAttributes = voidNode.getAttributes();

            if (voidNode.getNodeName()
                        .equalsIgnoreCase(XMLDataAccessObject.VOID))
                {
                Node propertyNode = voidAttributes.getNamedItem(
                                            XMLDataAccessObject.PROPERTY);
                Field field = getField(instance.getClass(), 
                                       propertyNode.getNodeValue());

                NodeList fieldValues = voidNode.getChildNodes();

                for (int j = 0; j < fieldValues.getLength(); j++)
                    {
                    Node fieldValue = fieldValues.item(j);

                    if (fieldValue.getNodeName()
                                  .equalsIgnoreCase(XMLDataAccessObject.OBJECT))
                        {
                        try
                            {
                            System.out.println("node name:" + 
                                               fieldValue.getNodeName() + 
                                               " attr:" + 
                                               fieldValue.getAttributes());
                            field.setAccessible(true);

                            field.set(instance, getObjectInstance(fieldValue));
                            }
                        catch (IllegalAccessException iae)
                            {
                            iae.printStackTrace();
                            }
                        }
                    else if (fieldValue.getNodeName()
                                       .equalsIgnoreCase(XMLDataAccessObject.ARRAY))
                        {
                        addObjectArrayProperties(field, instance, fieldValue);
                        }
                    else
                        {
                        Node valueNode = fieldValue.getFirstChild();
                        FieldType fieldType = getFieldTypeMap()
                                                  .getFieldType(fieldValue.getNodeName());
                        fieldType.setPrimitiveValue(field, instance, valueNode);
                        }
                    }
                }
            }
        }

    // handle the array of object's state
    private void addObjectArrayProperties(Field field, Object instance, 
                                          Node arrayNode)
        {
        NodeList objectNodeList = arrayNode.getChildNodes();
        Object[] array = new Object[objectNodeList.getLength()];

        if (!field.getType().getComponentType().isPrimitive())
            {
            array = (Object[]) Array.newInstance(field.getType()
                                                      .getComponentType(), 
                                                 objectNodeList.getLength());
            }

        for (int i = 0; i < objectNodeList.getLength(); i++)
            {
            Node objectNode = objectNodeList.item(i);
            array[i] = null;

            if (objectNode.getNodeName()
                          .equalsIgnoreCase(XMLDataAccessObject.STRING))
                {
                array[i] = objectNode.getLastChild().getNodeValue();
                }
            else if (objectNode.getNodeName()
                               .equalsIgnoreCase(XMLDataAccessObject.NULL))
                {
                array[i] = null;
                }
            else if (getFieldTypeMap().containsField(objectNode.getNodeName()))
                {
                FieldType fieldType = getFieldTypeMap()
                                          .getFieldType(objectNode.getNodeName());
                fieldType.restorePrimitiveArray(field, instance, arrayNode);
                }
            else if (objectNode.getNodeName()
                               .equalsIgnoreCase(XMLDataAccessObject.OBJECT))
                {
                array[i] = getObjectInstance(objectNode);
                }
            }

        try
            {
            field.set(instance, array);
            }
        catch (IllegalAccessException iae)
            {
            iae.printStackTrace();
            }
        catch (IllegalArgumentException iae2)
            {
            }
        }

    // retrun the field by name
    private Field getField(Class type, String name)
        {
        Field returnField = null;
        Field[] fields = type.getDeclaredFields();

        for (int i = 0; i < fields.length; i++)
            {
            Field field = fields[i];
            field.setAccessible(true);

            if (field.getName().equals(name))
                {
                returnField = field;
                i = fields.length;
                }
            }

        if (returnField == null)
            {
            returnField = getField(type.getSuperclass(), name);
            }

        return returnField;
        }

    // check if the the Element is support in the xml file.
    private boolean supported(NamedNodeMap attributes, String name)
        {
        return (attributes.getNamedItem(name) instanceof Node);
        }
    }